from typing import Any, Optional, ClassVar, Callable, Awaitable
from dataclasses import dataclass
from http import HTTPStatus
import traceback

from starlette.requests import Request
from starlette.responses import Response, JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint

@dataclass
class HTTPError(Exception):
	desc: str
	data: Optional[Any] = None
	code: ClassVar[int]

@dataclass
class BadRequest(HTTPError):
	code = 400

@dataclass
class Unauthorised(HTTPError):
	code = 401

@dataclass
class Forbidden(HTTPError):
	code = 403

@dataclass
class Conflict(HTTPError):
	code = 409

def error_response(code: int, desc: Optional[str], data: Optional[Any]) -> JSONResponse:
	return JSONResponse({
		'error_type': HTTPStatus(code).phrase,
		'error_description': desc,
		'error_data': data,
	}, code)

class JSONErrorMiddleware(BaseHTTPMiddleware):
	async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
		# TODO: logging
		try:
			response = await call_next(request)
			return response
		except HTTPError as e:
			return error_response(e.code, e.desc, e.data)
		except Exception as e:
			traceback.print_exc()
			return error_response(500, None, None)



js = '''
window.onload = function() {
	const ui = SwaggerUIBundle({
		url: "/openapi/schema",
		dom_id: '#swagger-ui',
		deepLinking: true,
		presets: [
			SwaggerUIBundle.presets.apis,
			SwaggerUIStandalonePreset,
		],
		plugins: [
			SwaggerUIBundle.plugins.DownloadUrl,
		],
		layout: "BaseLayout",
	})
	window.ui = ui
}
'''

def build_swagger_ui(title: str) -> str:
	return f'''
	<html>
	<head>
		<meta charset="UTF-8">
		<title>{title}</title>
		<link rel="stylesheet" type="text/css" href="https://unpkg.com/swagger-ui-dist@3.12.1/swagger-ui.css">
	</head>
	<body>

	<div id="swagger-ui"></div>

	<script src="https://unpkg.com/swagger-ui-dist@3.12.1/swagger-ui-standalone-preset.js"></script>
	<script src="https://unpkg.com/swagger-ui-dist@3.12.1/swagger-ui-bundle.js"></script>

	<script>
	{js}
	</script>
	</body>
	</html>
	'''



from typing import Any, Optional, Dict, List

from starlette.applications import Starlette
from starlette.responses import JSONResponse, HTMLResponse
from starlette.routing import Route, Mount
from starlette.middleware import Middleware
from starlette.types import ASGIApp
import uvicorn

from .route_def import make_route_def
from .schemas import build_schemas
from .swagger import build_swagger_ui
from .endpoint import make_endpoint
from .errors import JSONErrorMiddleware
from .auth import AuthBackend, AuthMiddleware

def create_app(srv: Any, name: Optional[str] = None, auth_backend: Optional[AuthBackend] = None) -> ASGIApp:
	name = name or type(srv).__name__
	route_defs = [ make_route_def(getattr(srv, m)) for m in dir(srv) if not m.startswith('_') and m not in { 'startup', 'shutdown' } ]

	openapi_schemas = build_schemas(route_defs, name, auth_backend)
	swagger_ui = build_swagger_ui(name)

	startup_funcs = []
	shutdown_funcs = []

	def _attach_lifecyle_methods(obj: Any) -> None:
		if startup_func := getattr(obj, 'startup', None):
			startup_funcs.append(startup_func)
		if shutdown_func := getattr(obj, 'shutdown', None):
			shutdown_funcs.append(shutdown_func)
	
	_attach_lifecyle_methods(srv)

	middleware: Any = [ Middleware(JSONErrorMiddleware) ]
	if auth_backend:
		middleware.append(Middleware(AuthMiddleware, auth_backend = auth_backend))
		_attach_lifecyle_methods(auth_backend)

	core_app = Starlette(
		routes = [
			Route(r.path, make_endpoint(r), methods = [ r.http_method.upper() ])
			for r in route_defs
		],
		middleware = middleware,
	)

	openapi_app = Starlette(
		routes = [
			Route('/swagger', lambda _: HTMLResponse(swagger_ui)),
			Route('/schema', lambda _: JSONResponse(openapi_schemas)),
		],
	)

	return Starlette(routes = [
		Mount('/openapi', openapi_app),
		Mount('/', core_app),
	], on_startup = startup_funcs, on_shutdown = shutdown_funcs)

def run_app(app: ASGIApp, port: int = 80, host: str = "0.0.0.0") -> None:
	uvicorn.run(app, host = host, port = port)
